/*

Licensed under the GNU AFFERO GENERAL PUBLIC LICENSE Version 3 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/agpl-3.0.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package cronjob

import (
	"errors"

	batchv1beta1 "k8s.io/api/batch/v1beta1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"

	"k8s.libre.sh/interfaces"
	"k8s.libre.sh/meta"
	"k8s.libre.sh/objects/job"
)

// +kubebuilder:object:generate=true
type CronJob struct {
	// Standard object metadata.
	// +optional
	*meta.ObjectMeta `json:"meta,omitempty"`
	// Specification of the desired behavior of the Job.
	// +optional
	*CronJobSpec `json:",inline"`
}

// +kubebuilder:object:generate=true
type CronJobSpec struct {
	// The schedule in Cron format, see https://en.wikipedia.org/wiki/Cron.
	Schedule string `json:"schedule" protobuf:"bytes,1,opt,name=schedule"`

	// Optional deadline in seconds for starting the job if it misses scheduled
	// time for any reason.  Missed jobs executions will be counted as failed ones.
	// +optional
	StartingDeadlineSeconds *int64 `json:"startingDeadlineSeconds,omitempty" protobuf:"varint,2,opt,name=startingDeadlineSeconds"`

	// Specifies how to treat concurrent executions of a Job.
	// Valid values are:
	// - "Allow" (default): allows CronJobs to run concurrently;
	// - "Forbid": forbids concurrent runs, skipping next run if previous run hasn't finished yet;
	// - "Replace": cancels currently running job and replaces it with a new one
	// +optional
	ConcurrencyPolicy batchv1beta1.ConcurrencyPolicy `json:"concurrencyPolicy,omitempty" protobuf:"bytes,3,opt,name=concurrencyPolicy,casttype=ConcurrencyPolicy"`

	// This flag tells the controller to suspend subsequent executions, it does
	// not apply to already started executions.  Defaults to false.
	// +optional
	Suspend *bool `json:"suspend,omitempty" protobuf:"varint,4,opt,name=suspend"`

	// The number of successful finished jobs to retain.
	// This is a pointer to distinguish between explicit zero and not specified.
	// Defaults to 3.
	// +optional
	SuccessfulJobsHistoryLimit *int32 `json:"successfulJobsHistoryLimit,omitempty" protobuf:"varint,6,opt,name=successfulJobsHistoryLimit"`

	// The number of failed finished jobs to retain.
	// This is a pointer to distinguish between explicit zero and not specified.
	// Defaults to 1.
	// +optional
	FailedJobsHistoryLimit *int32 `json:"failedJobsHistoryLimit,omitempty" protobuf:"varint,7,opt,name=failedJobsHistoryLimit"`
	// Job describes the pod that will be created.
	*job.Job `json:",inline"`
}

func (spec *CronJobSpec) GetSuccessfulJobsHistoryLimit() *int32 {
	return spec.SuccessfulJobsHistoryLimit
}

func (spec *CronJobSpec) GetFailedJobsHistoryLimit() *int32 {
	return spec.FailedJobsHistoryLimit
}

func (spec *CronJobSpec) GetConcurrencyPolicy() batchv1beta1.ConcurrencyPolicy {
	return spec.ConcurrencyPolicy
}

func (spec *CronJobSpec) GetStartingDeadlineSeconds() *int64 {
	return spec.StartingDeadlineSeconds
}

func (spec *CronJobSpec) GetSchedule() string {
	return spec.Schedule
}

func (o *CronJob) GetObject() interfaces.Object {
	obj := &batchv1beta1.CronJob{
		TypeMeta: metav1.TypeMeta{
			Kind:       "CronJob",
			APIVersion: "batch/v1beta1",
		},
	}

	obj.SetName(o.GetName())
	obj.SetNamespace(o.GetNamespace())

	return obj
}

func (o *CronJob) Mutate(obj interfaces.Object) error {

	cron, ok := obj.(*batchv1beta1.CronJob)
	if !ok {
		return errors.New("object is not a CronJob")
	}

	// TODO TO FIX merge meta
	o.Pod.ObjectMeta = o.ObjectMeta

	err := MutateCronJob(o, cron)
	if err != nil {
		return err
	}

	err = meta.MutateMeta(o.ObjectMeta, obj)
	if err != nil {
		return err
	}

	cron.Spec.JobTemplate.ObjectMeta.Name = cron.ObjectMeta.Name
	cron.Spec.JobTemplate.ObjectMeta.Labels = o.ObjectMeta.Labels

	return nil
}

func (o *CronJob) Init() {
	if o.ObjectMeta == nil {
		o.ObjectMeta = &meta.ObjectMeta{}
	}

	if o.ObjectMeta.Labels == nil {
		o.ObjectMeta.Labels = make(map[string]string)
	}

	if o.CronJobSpec == nil {
		o.CronJobSpec = &CronJobSpec{}
	}

	if o.Job == nil {
		o.Job = &job.Job{}
	}

	o.Job.Init()

}

func (o *CronJob) GetObjectKind() string  { return "CronJob" }
func (o *CronJob) GetObjectGroup() string { return "batchv1beta1" }
